import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

/*
 * This program illustrates the use of locks (Java ReentrantLock class) and
 * condition variables (Java Condition class) to solve the Bounded Buffer
 * problem. 
 */

public class BoundedBuffer {
	
	// Size of the bounded buffer.
	private final int size;
	// Number of empty slots in the buffer.  The number of full slots is
	// (size - empty).
	private int empty;
	// Index of the next empty slot.
	private int in;
	// Index of the next full slot.
	private int out;
	
	// Lock for controlling access to the buffer.
	private final ReentrantLock mutex;
	// Waiting for a full buffer slot.
	private final Condition fullCond;
	// Waiting for an empty buffer slot.
	private final Condition emptyCond;
	
	// The buffer.
	private final String[] data;
	
	public BoundedBuffer(int size) {
		this.size = size;
		empty = size;
		mutex = new ReentrantLock();
		fullCond = mutex.newCondition();
		emptyCond = mutex.newCondition();
		data = new String[size];
		in = 0;
		out = 0;
		
		// Create runnable objects, bind them to threads, and start the
		// threads.
		new Thread(producer(0, 1000)).start();
		new Thread(producer(1, 1000)).start();
		new Thread(consumer(2, 1000)).start();
		new Thread(consumer(3, 1000)).start();
		
	}

	private Runnable producer(int id, int delay) {
		
		return () -> {
			int count = 0;
			
			while (true) {
				String s = id + " produces " + count++;
				
				mutex.lock();
				
				// If there are no empty buffers, wait.
				// It is CRUCIAL to ALWAYS recheck the condition when
				// notified.
				while (empty == 0) {
					try {
						emptyCond.await();
					} catch (InterruptedException e) {
					}
				}

				// Produce into the buffer.
				data[in] = s;
				in = (in + 1) % size;
				empty--;

				// Signal a consumer that there a full slot in the buffer.
				fullCond.signal();
				mutex.unlock();
				System.out.println(s);
				// Take a nap.
				try {
					Thread.sleep(delay);
				} catch (InterruptedException e) {
				}
			}
		};
	}

	private Runnable consumer(int id, int delay) {

		return () -> {
			int count = 0;

			while (true) {
				mutex.lock();

				// If there are no full buffers, wait.
				// It is CRUCIAL to ALWAYS recheck the condition when
				// notified.
				while ((size - empty) == 0) {
					try {
						fullCond.await();
					} catch (InterruptedException e) {
					}
				}

				// Consume from the buffer.
				String s = data[out];
				out = (out + 1) % size;
				empty++;

				// Signal a producer that there an empty slot in the buffer.
				emptyCond.signal();
				mutex.unlock();
				
				System.out.println(id + " consumes " + count++ + ": " + s);
				// Take a nap.
				try {
					Thread.sleep(delay);
				} catch (InterruptedException e) {
				}
			}

		};
	}
	
	public static void main(String[] args) {
		new BoundedBuffer(5);
	}

}
